@@ -23,8 +23,6 @@ except ImportError:
     "some functions dealing with the binary file (filtering, whitening...) will not work."))
 
 from npyx.utils import npa, read_pyfile, list_files
-from npyx.preprocess import apply_filter, bandpass_filter, whitening, approximated_whitening_matrix, med_substract,\
-                            gpufilter, cu_median, adc_realign, kfilt
 
 import json
 
@@ -533,27 +531,36 @@ def get_npix_sync(dp, output_binary = False, filt_key='highpass', unit='seconds'
         return onsets,offsets
 
 def extract_rawChunk(dp, times, channels=np.arange(384), filt_key='highpass', save=0,
-                     whiten=0, med_sub=0, hpfilt=0, hpfiltf=300, nRangeWhiten=None, nRangeMedSub=None,
-                     ignore_ks_chanfilt=0, center_chans_on_0=False, verbose=False, scale=True, again=False):
+                     whiten=0, med_sub=0, hpfilt=0, hpfiltf=300, filter_forward=True, filter_backward=True,
+                     nRangeWhiten=None, nRangeMedSub=None, use_ks_w_matrix=True,
+                     ignore_ks_chanfilt=True, center_chans_on_0=False, verbose=False, scale=True, again=False):
     '''Function to extract a chunk of raw data on a given range of channels on a given time window.
-    ## PARAMETERS
+    Parameters:
     - dp: datapath to folder with binary path (files must ends in .bin, typically ap.bin)
     - times: list of boundaries of the time window, in seconds [t1, t2].
     - channels (default: np.arange(384)): list of channels of interest, in 0 indexed integers [c1, c2, c3...]
     - filt_key: 'ap' or 'lf', whether to exxtract from the high-pass or low-pass filtered binary file
     - save (default 0): save the raw chunk in the bdp directory as '{bdp}_t1-t2_c1-c2.npy'
-    - whiten: whether to whiten the data across channels. If nRangeWhiten is not None, whitening matrix is computed with the nRangeWhiten closest channels.
-    - med_sub: whether to median-subtract the data across channels. If nRangeMedSub is not none, median of each channel is computed using the nRangeMedSub closest channels.
-    - hpfilt: whether to high-pass filter the data, using a 3 nodes butterworth filter of cutoff frequency hpfiltf.
+    - whiten: whether to whiten the data across channels. If nRangeWhiten is not None,
+              whitening matrix is computed with the nRangeWhiten closest channels.
+    - med_sub: whether to median-subtract the data across channels. If nRangeMedSub is not none,
+               median of each channel is computed using the nRangeMedSub closest channels.
+    - hpfilt: whether to high-pass filter the data, using a 3 nodes butterworth filter of cutoff frequency hpfiltf, bidirectionally.
     - hpfiltf: see hpfilt
+    - filter_forward: bool, filter the data forward (also set filter_backward to True for bidirectional filtering)
+    - filter_backward: bool, filter the data backward (also set filter_forward to True for bidirectional filtering)
     - nRangeWhiten: int, see whiten.
     - nRangeMedSub: int, see med_sub.
-    - ignore_ks_chanfilt: whether to ignore the filtering made by kilosort, which only uses channels with average events rate > ops.minfr to spike sort. | Default False
+    - use_ks_w_matrix: bool, whether to use kilosort's original whitening matrix to perform the whitening
+                     (rather than recomputing it from the data at hand)
+    - ignore_ks_chanfilt: whether to ignore the filtering made by kilosort,
+                          which only uses channels with average events rate > ops.minfr to spike sort.
     - scale: A boolean variable specifying whether we should convert the resulting raw
              A2D samples to uV. Defaults to True
-    ## RETURNS
-    rawChunk: numpy array of shape ((c2-c1), (t2-t1)*fs).
-    rawChunk[0,:] is channel 0; rawChunk[1,:] is channel 1, etc.
+    Returns:
+    - rawChunk: numpy array of shape ((c2-c1), (t2-t1)*fs).
+                rawChunk[0,:] is channel 0; rawChunk[1,:] is channel 1, etc.
+                dtype: int16 if scale=False, float64 if scale=True.
     '''
     # Find binary file
     dp = Path(dp)
@@ -570,8 +577,7 @@ def extract_rawChunk(dp, times, channels=np.arange(384), filt_key='highpass', sa
     assert times[1]<meta['recording_length_seconds']
 
     # Format inputs
-    cm=chan_map(dp, probe_version='local'); assert cm.shape[0]<=Nchans-1
-    if not ignore_ks_chanfilt: channels=assert_chan_in_dataset(dp, channels) # index out of 384, should remain the same because rc initial shape is 384!
+    channels=assert_chan_in_dataset(dp, channels, ignore_ks_chanfilt)
     t1, t2 = int(np.round(times[0]*fs)), int(np.round(times[1]*fs))
     if whiten:
         if t1<whitenpad:
@@ -580,7 +586,9 @@ def extract_rawChunk(dp, times, channels=np.arange(384), filt_key='highpass', sa
             times[0] = t1/30000
         t1, t2 = t1-whitenpad, t2+whitenpad
     bn = os.path.basename(fname) # binary name
-    rcn = f'{bn}_t{times[0]}-{times[1]}_ch{channels[0]}-{channels[-1]}_{whiten}_{med_sub}_{scale}.npy' # raw chunk name
+    rcn = (f"{bn}_t{times[0]}-{times[1]}_ch{channels[0]}-{channels[-1]}"
+          f"_{whiten}{nRangeWhiten}_{med_sub}{nRangeMedSub}_{hpfilt}{hpfiltf}"
+          f"_{ignore_ks_chanfilt}_{center_chans_on_0}_{scale}.npy") # raw chunk name
     rcp = get_npyx_memory(dp) / rcn
 
     if os.path.isfile(rcp) and not again:
@@ -613,31 +621,45 @@ def extract_rawChunk(dp, times, channels=np.arange(384), filt_key='highpass', sa
     assert len(bData)%2==0
     rc = np.frombuffer(bData, dtype=np.int16) # 16bits decoding
     rc = rc.reshape((int(t2-t1), Nchans)).T
-
-    # Align signal on each channel
-    if center_chans_on_0:
-        rc=rc-np.median(rc[:,:10],axis=1)[:,np.newaxis]
+    rc = rc[:-1,:] # remove sync channel
 
     # Median subtraction = CAR
     if med_sub:
         rc=med_substract(rc, 0, nRange=nRangeMedSub)
 
-    # Highpass filter with a 3rd order butterworth filter, like in kilosort2
+    # get the right channels range,
+    # after median subtraction
+    if (whiten and use_ks_w_matrix) or not whiten:
+        # saves computation time to preselect channels
+        rc = rc[channels, :]
+
+    # Highpass filter with a 3rd order butterworth filter
     if hpfilt:
-        rc=apply_filter(rc, bandpass_filter(rate=fs, low=None, high=hpfiltf, order=3), axis=1)
+        rc_t = np.ascontiguousarray(rc.T)
+        rc_t = cp.asarray(rc_t)
+        rc = gpufilter(rc_t, fs=fs, fslow=None, fshigh=hpfiltf, order=3,
+             car=False, forward=filter_forward, backward=filter_backward, ret_numpy=True).T
 
     # Whiten data
     if whiten:
-        rc=whitening(rc, nRange=nRangeWhiten)
+        channels_mask = np.isin(np.arange(Nchans-1), channels)
+        rc=whitening(rc, nRangeWhiten, use_ks_w_matrix, dp, channels_mask)
         rc=rc[:, whitenpad:-whitenpad]
 
     # get the right channels range, AFTER WHITENING
-    rc = rc[channels, :]
+    if whiten and not use_ks_w_matrix:
+        rc = rc[channels, :]
+
     # Scale data
     if scale:
-        rc = rc*meta['bit_uV_conv_factor'] # convert into uV
+        rc = rc * meta['bit_uV_conv_factor'] # convert into uV
 
-    # convert from cupy to numpy array
+    # Align signal on each channel, option convenient for plotting
+    if center_chans_on_0:
+        rc=rc-np.median(rc[:,:10],axis=1)[:,np.newaxis]
+
+    # eventually convert from cupy to numpy array
+    # (necessary if whitened on GPU)
     if 'cp' in globals():
         rc = cp.asnumpy(rc)
 
@@ -646,9 +668,13 @@ def extract_rawChunk(dp, times, channels=np.arange(384), filt_key='highpass', sa
 
     return rc
 
-def assert_chan_in_dataset(dp, channels):
+def assert_chan_in_dataset(dp, channels, ignore_ks_chanfilt=False):
     channels = np.array(channels)
-    cm=chan_map(dp, probe_version='local')
+    if ignore_ks_chanfilt:
+        probe_version = read_metadata(dp)['probe_version']
+        cm=chan_map(dp, probe_version=probe_version)
+    else:
+        cm=chan_map(dp, probe_version='local')
     if not np.all(np.isin(channels, cm[:,0])):
         print(("WARNING Kilosort excluded some channels that you provided for analysis "
                "because they did not display enough threshold crossings! Jumping channels:"
@@ -659,7 +685,8 @@ def assert_chan_in_dataset(dp, channels):
 #%% Binary file filtering wrappers
 
 def preprocess_binary_file(dp=None, filt_key='ap', fname=None, target_dp=None, move_orig_data=True,
-                       ADC_realign = False, median_subtract=False, f_low=None, f_high=300, order=3,
+                       ADC_realign = False, median_subtract=True, f_low=None, f_high=300, order=3,
+                       filter_forward=True, filter_backward=False,
                        spatial_filt=False, whiten = False, whiten_range=32,
                        again_Wrot=False, verbose=False):
     """Creates a preprocessed copy of binary file at dp/fname_filtered.bin,
@@ -691,6 +718,8 @@ def preprocess_binary_file(dp=None, filt_key='ap', fname=None, target_dp=None, m
     - f_low: optional float, lowpass filter frequency
     - f_high: float, highpass filter frequency (necessary)
     - order: int, butterworth filter order (default 3)
+    - filter_forward: bool, filter the data forward (also set filter_backward to True for bidirectional filtering)
+    - filter_backward: bool, filter the data backward (also set filter_forward to True for bidirectional filtering)
     - spatial_filt: bool, whether to high pass filter across channels at 0.1 Hz
     - whiten: bool, whether to whiten across channels
     - verbose: bool, whether to print extra information
@@ -720,25 +749,10 @@ def preprocess_binary_file(dp=None, filt_key='ap', fname=None, target_dp=None, m
         target_dp = Path(target_dp)
     print(f"Preprocessing {fname}...")
 
-    filter_suffix = ""
-    message = ""
-    if ADC_realign:
-        filter_suffix+=f"_adcshift{ADC_realign}"
-        message+="    - shifting ADCs,\n"
-    if median_subtract:
-        filter_suffix+="_medsub"
-        message+="    - median subtraction (aka common average referencing CAR),\n"
-    filter_suffix+=f"_tempfilt{f_low}{f_high}"
-    low_s = 0 if f_low is None else f_low
-    message+=f"    - filtering in time (between {low_s} and {f_high} Hz),\n"
-    if whiten:
-        filter_suffix+=f"_whit{whiten}{whiten_range}"
-        message+=f"    - whitening (using {whiten_range} closest channels),\n"
-    if spatial_filt:
-        filter_suffix+=f"_spatfilt{spatial_filt}"
-        message+=f"    - filtering in space ({spatial_filt} 'Hz'),\n"
-    filtered_fname = str(fname.name)[:-7]+filter_suffix+".ap.bin"
-    message = message[:-2]+"."
+    filtered_fname, message = \
+        make_preprocessing_fname(fname, ADC_realign, median_subtract,
+                            f_low, f_high, filter_forward, filter_backward,
+                            whiten, whiten_range, spatial_filt)
     print(message)
 
     # fetch metadata
@@ -760,11 +774,12 @@ def preprocess_binary_file(dp=None, filt_key='ap', fname=None, target_dp=None, m
     offset = 0
     item_size = np.dtype(dtype).itemsize
     n_samples = (binary_byte_size - offset) // (item_size * n_channels)
-    memmap_f = np.memmap(fname, dtype=dtype, offset=offset, shape=(n_samples, n_channels), mode='r+')
+    memmap_f = np.memmap(fname, dtype=dtype, offset=offset, shape=(n_samples, n_channels), mode='r')
 
 
     # fetch whitening matrix (estimate covariance over a few batches)
     if whiten:
+        raise ImplementationError("Whitening here is still experimental - cannot be used for now.")
         Wrot_path = dp / 'whitening_matrix.npy'
         Wrot = approximated_whitening_matrix(memmap_f, Wrot_path, whiten_range,
             NT, Nbatch, NTbuff, ntb, nSkipCov, n_channels, channels_to_process,
@@ -777,6 +792,8 @@ def preprocess_binary_file(dp=None, filt_key='ap', fname=None, target_dp=None, m
     w_edge = cp.linspace(0,1,ntb).reshape(-1, 1) # weights to combine data batches at the edge
     buff_prev = cp.zeros((ntb, n_channels-1), dtype=np.int32)
     last_batch=False
+    assert not (target_dp / filtered_fname).exists(),\
+        f"WARNING file {target_dp / filtered_fname} exists already - to process again, delete or move it."
     with open(target_dp / filtered_fname, 'wb') as fw:  # open for writing processed data
         for ibatch in tqdm(range(Nbatch), desc="Preprocessing"):
             # we'll create a binary file of batches of NT samples, which overlap consecutively
@@ -816,7 +833,7 @@ def preprocess_binary_file(dp=None, filt_key='ap', fname=None, target_dp=None, m
 
             # CAR (optional) -> temporal filtering -> weight to combine edges -> unpadding
             batch = gpufilter(batch, fs=fs, fshigh=f_high, fslow=f_low, order=order,
-                              car=median_subtract, bidirectional=False)
+                              car=median_subtract, forward=filter_forward, backward=filter_backward)
             assert batch.flags.c_contiguous # check that ordering is still C, not F
             batch[ntb:2*ntb] = w_edge * batch[ntb:2*ntb] + (1 - w_edge) * buff_prev
             buff_prev = batch[NT + ntb: NT + 2*ntb]
@@ -827,9 +844,9 @@ def preprocess_binary_file(dp=None, filt_key='ap', fname=None, target_dp=None, m
                 batch = kfilt(batch.T, butter_kwargs = {'N': 3, 'Wn': 0.1, 'btype': 'highpass'}).T
 
             # whiten the data and scale by 200 for int16 range
-            if whiten:
-                print("Whitening not implemented yet.")
-                batch = cp.dot(batch, Wrot)
+            ## TODO implement whitening here
+            # if whiten:
+            #     batch = cp.dot(batch, Wrot)
             
             assert batch.flags.c_contiguous  # check that ordering is still C, not F
             if batch.shape[0] != NT:
@@ -852,6 +869,11 @@ def preprocess_binary_file(dp=None, filt_key='ap', fname=None, target_dp=None, m
                 print(f"{(target_dp/filtered_fname).stat().st_size} total, {rebuilt_batch.size * 2} bytes written to file {datcpu.shape} array size")
             datcpu.tofile(fw)
         if verbose: print(f"{(target_dp/filtered_fname).stat().st_size} total")
+    
+    # Apparently Windows is unhappy if memory mapped files
+    # aren't explicitely closed if trying to rename/move them
+    # so... close the memory mapped file o_o
+    memmap_f._mmap.close()
 
     # Finally, if everything ran smoothly,
     # move original binary to new directory
@@ -867,9 +889,38 @@ def preprocess_binary_file(dp=None, filt_key='ap', fname=None, target_dp=None, m
                 shutil.copy(dp/'channel_map.npy', orig_dp/'channel_map.npy')
             if (dp/'channel_positions.npy').exists():
                 shutil.copy(dp/'channel_positions.npy', orig_dp/'channel_positions.npy')
+            if (dp/'whitening_mat.npy').exists():
+                shutil.copy(dp/'whitening_mat.npy', orig_dp/'whitening_mat.npy')
 
     return target_dp/filtered_fname
 
+def make_preprocessing_fname(fname, ADC_realign, median_subtract,
+                            f_low, f_high, filter_forward, filter_backward,
+                            whiten, whiten_range, spatial_filt):
+
+    filter_suffix = ""
+    message = ""
+    if ADC_realign:
+        filter_suffix+=f"_adcshift{ADC_realign}"
+        message+="    - shifting ADCs,\n"
+    if median_subtract:
+        filter_suffix+="_medsub"
+        message+="    - median subtraction (aka common average referencing CAR),\n"
+    filter_suffix+=f"_tempfilt{f_low}{f_high}{filter_forward}{filter_backward}"
+    low_s = 0 if f_low is None else f_low
+    message+=(f"    - filtering in time (between {low_s} and {f_high} Hz)"
+              f" forward:{filter_forward}, backward:{filter_backward},\n")
+    if whiten:
+        filter_suffix+=f"_whit{whiten}{whiten_range}"
+        message+=f"    - whitening (using {whiten_range} closest channels),\n"
+    if spatial_filt:
+        filter_suffix+=f"_spatfilt{spatial_filt}"
+        message+=f"    - filtering in space ({spatial_filt} 'Hz'),\n"
+    filtered_fname = str(fname.name)[:-7]+filter_suffix+".ap.bin"
+    message = message[:-2]+"."
+
+    return filtered_fname, message
+
 #%% paqIO file loading utilities
 
 def paq_read(file_path):
@@ -949,190 +1000,9 @@ def paq_read(file_path):
             "units": units,
             "rate": rate}
 
-#%% I/O array functions from phy
-
-# def _start_stop(item):
-#     """Find the start and stop indices of a __getitem__ item.
-
-#     This is used only by ConcatenatedArrays.
-
-#     Only two cases are supported currently:
-
-#     * Single integer.
-#     * Contiguous slice in the first dimension only.
-
-#     """
-#     if isinstance(item, tuple):
-#         item = item[0]
-#     if isinstance(item, slice):
-#         # Slice.
-#         if item.step not in (None, 1):
-#             raise NotImplementedError()
-#         return item.start, item.stop
-#     elif isinstance(item, (list, np.ndarray)):
-#         # List or array of indices.
-#         return np.min(item), np.max(item)
-#     else:
-#         # Integer.
-#         return item, item + 1
-
-# def _fill_index(arr, item):
-#     if isinstance(item, tuple):
-#         item = (slice(None, None, None),) + item[1:]
-#         return arr[item]
-#     else:
-#         return arr
-
-# class ConcatenatedArrays(object):
-#     """This object represents a concatenation of several memory-mapped
-#     arrays. Coming from phy.io.array.py"""
-#     def __init__(self, arrs, cols=None, scaling=None):
-#         assert isinstance(arrs, list)
-#         self.arrs = arrs
-#         # Reordering of the columns.
-#         self.cols = cols
-#         self.offsets = np.concatenate([[0], np.cumsum([arr.shape[0]
-#                                                        for arr in arrs])],
-#                                       axis=0)
-#         self.dtype = arrs[0].dtype if arrs else None
-#         self.scaling = scaling
-
-#     @property
-#     def shape(self):
-#         if self.arrs[0].ndim == 1:
-#             return (self.offsets[-1],)
-#         ncols = (len(self.cols) if self.cols is not None
-#                  else self.arrs[0].shape[1])
-#         return (self.offsets[-1], ncols)
-
-#     def _get_recording(self, index):
-#         """Return the recording that contains a given index."""
-#         assert index >= 0
-#         recs = np.nonzero((index - self.offsets[:-1]) >= 0)[0]
-#         if len(recs) == 0:  # pragma: no cover
-#             # If the index is greater than the total size,
-#             # return the last recording.
-#             return len(self.arrs) - 1
-#         # Return the last recording such that the index is greater than
-#         # its offset.
-#         return recs[-1]
-
-#     def _get(self, item):
-#         cols = self.cols if self.cols is not None else slice(None, None, None)
-#         # Get the start and stop indices of the requested item.
-#         start, stop = _start_stop(item)
-#         # Return the concatenation of all arrays.
-#         if start is None and stop is None:
-#             return np.concatenate(self.arrs, axis=0)[..., cols]
-#         if start is None:
-#             start = 0
-#         if stop is None:
-#             stop = self.offsets[-1]
-#         if stop < 0:
-#             stop = self.offsets[-1] + stop
-#         # Get the recording indices of the first and last item.
-#         rec_start = self._get_recording(start)
-#         rec_stop = self._get_recording(stop)
-#         assert 0 <= rec_start <= rec_stop < len(self.arrs)
-#         # Find the start and stop relative to the arrays.
-#         start_rel = start - self.offsets[rec_start]
-#         stop_rel = stop - self.offsets[rec_stop]
-#         # Single array case.
-#         if rec_start == rec_stop:
-#             # Apply the rest of the index.
-#             out = _fill_index(self.arrs[rec_start][start_rel:stop_rel], item)
-#             out = out[..., cols]
-#             return out
-#         chunk_start = self.arrs[rec_start][start_rel:]
-#         chunk_stop = self.arrs[rec_stop][:stop_rel]
-#         # Concatenate all chunks.
-#         l = [chunk_start]
-#         if rec_stop - rec_start >= 2:
-#             print("Loading a full virtual array: this might be slow "
-#                         "and something might be wrong.")
-#             l += [self.arrs[r][...] for r in range(rec_start + 1,
-#                                                    rec_stop)]
-#         l += [chunk_stop]
-#         # Apply the rest of the index.
-#         return _fill_index(np.concatenate(l, axis=0), item)[..., cols]
-
-#     def __getitem__(self, item):
-#         out = self._get(item)
-#         assert out is not None
-#         if self.scaling is not None and self.scaling != 1:
-#             out = out * self.scaling
-#         return out
-
-#     def __len__(self):
-#         return self.shape[0]
-
-# def _pad(arr, n, dir='right'):
-#     """Pad an array with zeros along the first axis.
-
-#     Parameters
-#     ----------
-
-#     n : int
-#         Size of the returned array in the first axis.
-#     dir : str
-#         Direction of the padding. Must be one 'left' or 'right'.
-
-#     """
-#     assert dir in ('left', 'right')
-#     if n < 0:
-#         raise ValueError("'n' must be positive: {0}.".format(n))
-#     elif n == 0:
-#         return np.zeros((0,) + arr.shape[1:], dtype=arr.dtype)
-#     n_arr = arr.shape[0]
-#     shape = (n,) + arr.shape[1:]
-#     if n_arr == n:
-#         assert arr.shape == shape
-#         return arr
-#     elif n_arr < n:
-#         out = np.zeros(shape, dtype=arr.dtype)
-#         if dir == 'left':
-#             out[-n_arr:, ...] = arr
-#         elif dir == 'right':
-#             out[:n_arr, ...] = arr
-#         assert out.shape == shape
-#         return out
-#     else:
-#         if dir == 'left':
-#             out = arr[-n:, ...]
-#         elif dir == 'right':
-#             out = arr[:n, ...]
-#         assert out.shape == shape
-#         return out
-
-# def _range_from_slice(myslice, start=None, stop=None, step=None, length=None):
-#     """Convert a slice to an array of integers."""
-#     assert isinstance(myslice, slice)
-#     # Find 'step'.
-#     step = myslice.step if myslice.step is not None else step
-#     if step is None:
-#         step = 1
-#     # Find 'start'.
-#     start = myslice.start if myslice.start is not None else start
-#     if start is None:
-#         start = 0
-#     # Find 'stop' as a function of length if 'stop' is unspecified.
-#     stop = myslice.stop if myslice.stop is not None else stop
-#     if length is not None:
-#         stop_inferred = floor(start + step * length)
-#         if stop is not None and stop < stop_inferred:
-#             raise ValueError("'stop' ({stop}) and ".format(stop=stop) +
-#                              "'length' ({length}) ".format(length=length) +
-#                              "are not compatible.")
-#         stop = stop_inferred
-#     if stop is None and length is None:
-#         raise ValueError("'stop' and 'length' cannot be both unspecified.")
-#     myrange = np.arange(start, stop, step)
-#     # Check the length if it was specified.
-#     if length is not None:
-#         assert len(myrange) == length
-#     return myrange
-
-
+class ImplementationError(Exception):
+    pass
 
 from npyx.gl import assert_multi, get_ds_table, get_npyx_memory
-#
+from npyx.preprocess import whitening, approximated_whitening_matrix, med_substract,\
+                            gpufilter, adc_realign, kfilt
\ No newline at end of file
