@@ -26,7 +26,8 @@ from tqdm.auto import tqdm
 
 #%% Whitening
 
-def whitening(x, nRange=None):
+def whitening(x, nRange=None, use_ks_matrix=True,
+             dp=None, channels_mask=None):
     '''
     Whitens along axis 0.
     For instance, time should be axis 1 and channels axis 0 to whiten across channels.
@@ -36,39 +37,125 @@ def whitening(x, nRange=None):
         - x: 2D array, axis 1 spans time, axis 0 observations (e.g. channel).
              Multiplying by the whitening matrix whitens across obervations.
         - nRange: if integer, number of channels to locally compute whitening filter (more robust to noise) | Default None
+        - use_ks_matrix: bool, whether to use kilosort's original whitening matrix to perform the whitening
+                     (rather than recomputing it from the data at hand)
+        - dp: str/Path, kilosort path with whitening matrix whiten_mat.npy
+        - channels_mask: boolean array, matches channels to whiten (important to save memory on large arrays)
     '''
     assert x.shape[1]>=x.shape[0]
+
     # Compute whitening matrix
-    w=whitening_matrix(x, nRange=nRange)
-    # Whiten
+    if use_ks_matrix:
+        w, unprocessed_channels = whitening_matrix(x, nRange, use_ks_matrix, dp)
+        
+        if channels_mask is None:
+            unprocessed_traces = x[unprocessed_channels,:]
+            x = x[~unprocessed_channels,:]
+        else:
+            # only subselect some channels for whitening if instructed
+            # can only work when loading precomputed whitening matrix
+            assert np.sum(channels_mask) == x.shape[0],\
+                "WARNING inconsistency between provided channel_mask and partial array to whiten!"
+            unprocessed_channels_masked = unprocessed_channels[channels_mask]
+            channels_mask = channels_mask[~unprocessed_channels]
+            channels_2D_mask = channels_mask[:,None] & channels_mask[None,:]
+            w = w[np.nonzero(channels_2D_mask)]
+            w = w.reshape((channels_mask.sum(), channels_mask.sum()))
+
+            unprocessed_traces = x[unprocessed_channels_masked,:]
+            x = x[~unprocessed_channels_masked, :]
+    else:
+        w = whitening_matrix(x, nRange, use_ks_matrix, dp)
+
+    
+    # Whiten and re-scale to match original microvolts
     x = cp.array(x)
     scales=(np.max(x, 1)-np.min(x, 1))
-    x=np.dot(x.T,w).T
+    x = np.dot(x.T,w).T
     W_scales=(np.max(x, 1)-np.min(x, 1))
     x=x*np.repeat((scales/W_scales).reshape(x.shape[0], 1), x.shape[1], axis=1)
 
+    # re-plug in unprocessed channels
+    if use_ks_matrix:
+        if channels_mask is not None:
+            x_full = cp.zeros((unprocessed_channels_masked.shape[0], x.shape[1]))
+            x_full[~unprocessed_channels_masked] = x
+            x_full[unprocessed_channels_masked] = unprocessed_traces
+        else:
+            x_full = cp.zeros((unprocessed_channels.shape[0], x.shape[1]))
+            x_full[~unprocessed_channels] = x
+            x_full[unprocessed_channels] = unprocessed_traces
+        
+        x = x_full
+    
+    if 'cp' in globals():
+        x = cp.asnumpy(x)
+
     return x
 
-def whitening_matrix(x, nRange=None):
+def whitening_matrix(x, nRange=None, use_ks_matrix=True, dp=None):
     """
-    wmat = whitening_matrix(dat, fudge=1e-18)
     Compute the whitening matrix using ZCA.
-        - dat is a matrix nsamples x nchannels
-    Apply using np.dot(dat,wmat)
-    Adapted from phy
+
     Parameters:
         - x: 2D array, axis 1 spans time, axis 0 observations (e.g. channel).
              Multiplying by the whitening matrix whitens across obervations.
         - epsilon: small value added to diagonal to regularize D
         - nRange: if integer, number of channels to locally compute whitening filter (more robust to noise) | Default None
+        - use_ks_matrix: bool, whether to return kilosort's original whitening matrix
+                     (rather than recomputing it from the data at hand)
+        - dp: str/Path, kilosort path with whitening matrix whiten_mat.npy
     """
     assert x.ndim == 2
     nrows, ncols = x.shape
 
+    if use_ks_matrix:
+        if nRange is not None:
+            print(("WARNING you instructed to use kilosort's original whitening matrix,"
+                  " so nRange is not taken into account (kilosort uses nRange=32 by default)"))
+        assert dp is not None, "You must provide a datapath when instructing to use kilosort's whitening matrix."
+        w, unprocessed_channels = load_ks_whitening_matrix(dp)
+        return cp.array(w), unprocessed_channels
+
     # get covariance matrix across rows (each row is an observation)
     x_cov = cp.cov(x, rowvar=1) 
     return cov_to_whitening_matrix(x_cov, nRange)
 
+def load_ks_whitening_matrix(dp, return_full=False):
+    """
+    Return kilosort whitening matrix
+    if return_full: also with missing channels (not preprocessed) replaced with 0s.
+
+    Parameters:
+    - dp: str/Path, datapath
+    - return_full: bool, whether to return full whitening matrix
+                   (adding arrays of 0s off-diagonal and 1 on-diagonal for channels skipped by kilosort)
+
+    Returns:
+    - kilosort whitening matrix (nchans, nchans) with nchans <= 384
+    - channels_mask: bool array, channels not processed by kilosort
+    """
+
+    probe_version = read_metadata(dp)['probe_version']
+    local_cm = chan_map(dp, probe_version='local')[:,0]
+    full_cm = chan_map(dp, probe_version=probe_version)[:,0]
+
+    channels_mask = np.isin(full_cm, local_cm)
+    channels_2D_mask = channels_mask[:,None] & channels_mask[None,:]
+
+    local_Wrot = np.load(Path(dp) / 'whitening_mat.npy')
+
+    if not return_full:
+        return local_Wrot, ~channels_mask
+
+    full_Wrot = np.zeros((full_cm.shape[0], full_cm.shape[0])) # * np.nan
+    full_Wrot[np.nonzero(channels_2D_mask)] = local_Wrot.ravel()
+    full_Wrot[np.nonzero(~channels_mask)[0], np.nonzero(~channels_mask)[0]] = 1
+
+    return full_Wrot, ~channels_mask
+
+
+
 def approximated_whitening_matrix(memmap_f, Wrot_path, whiten_range,
         NT, Nbatch, NTbuff, ntb, nSkipCov, n_channels, channels_to_process,
         f_high, fs, again=False, verbose=False):
@@ -284,16 +371,28 @@ def bandpass_filter(rate=None, low=None, high=None, order=1):
     elif high is not None:
         return sgnl.butter(order, high, 'highpass', fs=rate)
 
-def apply_filter(x, filt, axis=0):
-    """Apply a filter to an array."""
+def apply_filter(x, filt, axis=0, forward=True, backward=True):
+    """Apply a filter to an array, bidirectionally."""
     x = np.asarray(x)
     if x.shape[axis] == 0:
         return x
     b, a = filt
-    return sgnl.filtfilt(b, a, x, axis=axis)
+
+    if forward and backward:
+        # probably faster to use filtfilt than lfilter twice
+        return sgnl.filtfilt(b, a, x, axis=axis)
+
+    elif forward:
+        return sgnl.lfilter(b, a, x, axis=axis)
+
+    else:
+        assert backward and not forward # precaution
+        x = np.flip(x, axis)
+        x = sgnl.lfilter(b, a, x, axis=axis)
+        return np.flip(x, axis)
 
 def gpufilter(buff, fs=None, fslow=None, fshigh=None, order=3,
-             car=False, bidirectional=True):
+             car=False, forward=True, backward=True, ret_numpy=False):
     # filter this batch of data after common average referencing with the
     # median
     # buff is timepoints by channels
@@ -306,9 +405,11 @@ def gpufilter(buff, fs=None, fslow=None, fshigh=None, order=3,
     assert dataRAW.ndim == 2
     assert dataRAW.shape[0] > dataRAW.shape[1]
     assert dataRAW.ndim == 2
+    assert forward or backward, "You should either filter forward or backward."
 
     # subtract the mean from each channel
-    dataRAW = dataRAW - cp.mean(dataRAW, axis=0)  # subtract mean of each channel
+    # Maxime: I would use the median, but kilosort uses the mean
+    dataRAW = dataRAW - cp.mean(dataRAW, axis=0)
     assert dataRAW.ndim == 2
 
     # CAR, common average referencing by median
@@ -321,10 +422,17 @@ def gpufilter(buff, fs=None, fslow=None, fshigh=None, order=3,
 
     # next four lines should be equivalent to filtfilt (which cannot be
     # used because it requires float64)
-    datr = lfilter(*filter_params, dataRAW, axis=0)  # causal forward filter
-    if bidirectional:
-        datr = lfilter(*filter_params, datr, axis=0, reverse=True)  # backward
-    return datr
+    if forward:
+        dataRAW = lfilter(*filter_params, dataRAW, axis=0)  # causal forward filter
+    if backward:
+        # Maxime note: I do not understand why pykilosort folks
+        # did not do the same as MATLAB kilosort (i.e. simply revesing dataRAW)
+        dataRAW = lfilter(*filter_params, dataRAW, axis=0, reverse=True)  # backward
+    
+    if ret_numpy:
+        dataRAW = cp.asnumpy(dataRAW)
+
+    return dataRAW
 
 def get_filter_params(fs, fshigh=None, fslow=None, order=3):
     if fslow and fslow < fs / 2:
@@ -910,4 +1018,7 @@ def ns_optim_fft(ns):
     """
     p2, p3 = np.meshgrid(2 ** np.arange(25), 3 ** np.arange(15))
     sz = np.unique((p2 * p3).flatten())
-    return sz[np.searchsorted(sz, ns)]
\ No newline at end of file
+    return sz[np.searchsorted(sz, ns)]
+
+
+from npyx.inout import chan_map, read_metadata
\ No newline at end of file
