@@ -21,7 +21,8 @@ from math import ceil
 import matplotlib.pyplot as plt
 
 from npyx.utils import npa, split, xcorr_1d_loop
-from npyx.inout import read_metadata, get_binary_file_path, chan_map, whitening, bandpass_filter, apply_filter, med_substract
+from npyx.inout import read_metadata, get_binary_file_path, chan_map
+from npyx.preprocess import whitening, bandpass_filter, apply_filter, med_substract
 from npyx.gl import get_units, get_npyx_memory
 
 def wvf(dp, u=None, n_waveforms=100, t_waveforms=82, selection='regular', periods='all',
@@ -195,9 +196,10 @@ def wvf_dsmatch(dp, u, n_waveforms=100, t_waveforms=82, periods='all',
                 wvf_batch_size=10, ignore_nwvf=True, med_sub = False, spike_ids = None,
                 save=True, verbose=False, again=False,
                 whiten=False,  hpfilt=False, hpfiltf=300, nRangeWhiten=None, nRangeMedSub=None,
-                n_waves_used_for_matching = 50000, peakchan_allowed_range=10,
+                n_waves_used_for_matching = 5000, peakchan_allowed_range=6,
                 use_average_peakchan = False, max_allowed_amplitude = 1800, max_allowed_shift=3,
-                n_waves_to_average=5000, plot_debug=False, do_shift_match=True, n_waveforms_per_batch=10):
+                n_waves_to_average=800, plot_debug=False, do_shift_match=True, n_waveforms_per_batch=10,
+                subselect_max_template=False):
     """
     ********
     Extract the drift and shift matched mean waveforms of the specified unit.
@@ -254,19 +256,22 @@ def wvf_dsmatch(dp, u, n_waveforms=100, t_waveforms=82, periods='all',
         - nRangeMedSub:       int, number of channels to use to compute the local median. | Default None
         - ignore_ks_chanfilt: bool, whether to ignore kilosort channel filtering
                                     (if False, output shape will always be n_waveforms x t_waveforms x 384) | Default False
-        - n_waves_used_for_matching:   int, how many spikes to subsample to perform matching (default 50000 waveforms)
-        - peakchan_allowed_range: int (channel id), maximum allowed distance between original pek channel and ds-matched drift channel
+        - n_waves_used_for_matching:   int, how many spikes to subsample to perform matching (default 5000 waveforms)
+        - peakchan_allowed_range: int (channel id), maximum allowed distance between original peak channel and ds-matched drift channel
         - use_average_peakchan: bool, if True simply use the channel with highest average amplitude across spikes as peak channel
                                        instead of using the channel where the most spikes peak on
         - max_allowed_amplitude: float, maximum amplitude in uV (peak to trough) that a spike average can have to be considered (above, must be artefactual)
         - max_allowed_shift: int (samples), maximum allowed temporal shift during shift-matching (see Shift-matching explanation above)
-        - n_waves_to_average: int, maximum number of waveforms averaged together (5000 waveforms = 500 batches)
+        - n_waves_to_average: int, maximum number of waveforms averaged together (500 waveforms = 50 batches).
+                              Do not confuse with n_waves_used_for matching, which is the initial distribution of waveforms to pick from.
         - plot_debug: bool, whether to plot informative histograms displaying the distribution of peak channels (Z drift matching),
                       amplitudes on this peak channel (XY drift matching) and shifts (shift matching)
         - do_shift_match: bool, whether to perform shift matching
         - n_waveforms_per_batch: int, number of waveforms to use per batch for drift matching
                                  (in an ideal world 1, but too noisy - we assume that
                                   n_waveforms_per_batch consecutive waveforms have the same drift state)
+        - subselect_max_template: bool, whether to only use the kilosort template with the largest amount of spikes to compute the waveform
+                                  (less likely to average together waveforms looking different)
 
     Returns:
         - peak_dsmatched_waveform: (n_samples,) array (t_waveforms samples) storing the peak channel waveform
@@ -301,6 +306,16 @@ def wvf_dsmatch(dp, u, n_waveforms=100, t_waveforms=82, periods='all',
 
     ## Extract spike ids so we can extract consecutive waveforms
     spike_ids_all = ids(dp, u, periods=periods)
+    # make sure to only select waveforms from 1 cluster if there was a merge
+    # (arbitrary decision: the cluster with the largest amount of spikes)
+    if subselect_max_template:
+        spike_templates=np.load(Path(dp,'spike_templates.npy')).squeeze()[spike_ids_all]
+        template_ids, template_ns = np.unique(spike_templates, return_counts=True)
+        template_to_use = template_ids[np.argmax(template_ns)]
+        spike_ids_all = spike_ids_all[spike_templates==template_to_use]
+        if verbose: print((f"Used 1/{len(template_ids)} templates (had {template_ns[template_ids==template_to_use]} spikes, "
+                          f"others {template_ns[template_ids!=template_to_use]} respectively."))
+    # group in batches
     spike_ids_split_all = split(spike_ids_all, n_waveforms_per_batch, return_last = False).astype(np.int64)
     
     ## Subsample waveforms based on available RAM
@@ -309,6 +324,8 @@ def wvf_dsmatch(dp, u, n_waveforms=100, t_waveforms=82, periods='all',
     single_w_size = wvf(dp, None, t_waveforms=t_waveforms, spike_ids=[0]).nbytes
     max_n_waveforms = available_RAM//single_w_size-100 # -100 to be safe
     n_waves_used_for_matching = min(n_waves_used_for_matching, max_n_waveforms)
+    if n_waves_used_for_matching<1000 and verbose:
+        print(f"WARNING using less than 1000 waveforms for drift matching. This can lead to noisy output.")
     
     # now subsample ids based on the RAM-safe n_waves_used_for_matching
     if n_waves_used_for_matching<len(spike_ids_all):
@@ -337,7 +354,8 @@ def wvf_dsmatch(dp, u, n_waveforms=100, t_waveforms=82, periods='all',
     original_peak_chan = get_peak_chan(dp, u)
     c_left, c_right = max(0, original_peak_chan-peakchan_allowed_range), min(original_peak_chan+peakchan_allowed_range, mean_waves.shape[2])
     # calculate amplitudes ("peak-to-peak"), but ONLY using 2ms (-30,30) in the middle
-    t1, t2 = max(0,mean_waves.shape[1]//2-30), min(mean_waves.shape[1]//2+30, mean_waves.shape[1])
+    amp_t_span = 20 #samples
+    t1, t2 = max(0,mean_waves.shape[1]//2-amp_t_span), min(mean_waves.shape[1]//2+amp_t_span, mean_waves.shape[1])
     amplitudes = np.ptp(mean_waves[:,t1:t2,c_left:c_right], axis=1)
     batch_peak_channels = np.zeros(shape=(spike_ids_split_indices.shape[0], 3))
     batch_peak_channels[:,0] = spike_ids_split_indices # store batch indices (batch = averaged 10 spikes)
@@ -358,7 +376,8 @@ def wvf_dsmatch(dp, u, n_waveforms=100, t_waveforms=82, periods='all',
 
     if plot_debug:
         fig = hist_MB(batch_peak_channels[:,1], a=peak_channel-20, b=peak_channel+20, s=1,
-        title=f'Z drift matching:\ndistribution of peak channel across spike batches\n({n_waveforms_per_batch} spikes/batch - mode: chan {peak_channel})')
+        title=f'{u} Z drift matching:\ndistribution of peak channel across spike batches\n({n_waveforms_per_batch} spikes/batch - mode: chan {peak_channel})',
+        xlabel="Channel")
         ylim = fig.get_axes()[0].get_ylim()
         fig.get_axes()[0].plot([peak_channel,peak_channel], ylim, color='red', ls='--')
         fig.get_axes()[0].set_ylim(ylim)
@@ -377,7 +396,8 @@ def wvf_dsmatch(dp, u, n_waveforms=100, t_waveforms=82, periods='all',
         max_amp_hist = np.max(batch_peak_channels[:,2])
         max_amp_hist += 10-max_amp_hist%10
         nbatches_hist = batch_peak_channels.shape[0]
-        fig = hist_MB(batch_peak_channels[:,2], a=10, b=max_amp_hist, s=5, color='grey', alpha=0.7)
+        fig = hist_MB(batch_peak_channels[:,2], a=10, b=max_amp_hist, s=5, color='grey', alpha=0.7,
+                    xlabel="Amplitude (uV)")
 
     # if less than n_driftmatched_subset batches below 95th percentile,
     # use all up to n_driftmatched_subset batches
@@ -393,10 +413,9 @@ def wvf_dsmatch(dp, u, n_waveforms=100, t_waveforms=82, periods='all',
 
     if plot_debug:
         fig = hist_MB(batch_peak_channels[:,2], a=10, b=max_amp_hist, s=5, ax=fig.get_axes()[0], color='orange', alpha=0.7,
-        title=(f'XY drift matching:\ndistribution of amplitude on peak channel across spike batches\n'
+        title=(f'{u} XY drift matching:\ndistribution of amplitude on peak channel across spike batches\n'
                f'({n_waveforms_per_batch} spikes/batch - {batch_peak_channels.shape[0]}/{nbatches_hist} batches)'))
 
-
     #### shift matching ####
     # extract drift-matched raw waveforms
     dsmatch_batch_ids = batch_peak_channels[:,0].astype(np.int64)
@@ -425,9 +444,11 @@ def wvf_dsmatch(dp, u, n_waveforms=100, t_waveforms=82, periods='all',
 
     if plot_debug:
         if verbose: print(f'Total averaged waveform batches ({n_waveforms_per_batch}/batch) after drift-shift matching: {batch_peak_channels.shape[0]}')
-        fig = quickplot_n_waves(np.mean(mean_waves[np.random.randint(0, mean_waves.shape[0], batch_peak_channels.shape[0]),:,:], axis=0), '', peak_channel)
-        fig = quickplot_n_waves(np.mean(drift_matched_batches, axis=0), '', peak_channel, fig=fig)
-        fig = quickplot_n_waves(drift_shift_matched_mean, 'raw:blue\ndrift-matched:orange\ndrift-shift-matched:green', peak_channel, fig=fig)
+        wave_baseline_toplot = wvf(dp, u, t_waveforms=t_waveforms)
+        # mean_waves[np.random.randint(0, mean_waves.shape[0], batch_peak_channels.shape[0]),:,:]
+        fig = quickplot_n_waves(np.mean(wave_baseline_toplot, axis=0), '', peak_channel, color='k')
+        fig = quickplot_n_waves(np.mean(drift_matched_batches, axis=0), '', peak_channel, fig=fig, color='darkgreen')
+        fig = quickplot_n_waves(drift_shift_matched_mean, 'raw:black\ndrift-matched:green\ndrift-shift-matched:red', peak_channel, fig=fig, color='red')
         #breakpoint()
 
     return drift_shift_matched_mean_peak, drift_shift_matched_mean, drift_matched_spike_ids, peak_channel
@@ -454,7 +475,7 @@ def shift_match(waves, alignment_channel,
         - chan_range: int, range of channels around alignment channel used for template matching
                       (3 corresponds to 6 channels, 5 to 10 etc)
         - recenter_spikes: bool, whether to align the maximum of template to 0
-        - dynamic_template: bool, whether to update the template by averaging it with the aligned spike
+        - dynamic_template: bool, whether to update the template by averaging it with the aligned spike (actually bad idea)
         - max_shift_allowed: int, maximum shift allowed (half window) - other waveforms are discarded (if need to shift more, they must be way too noisy)
 
     Returns:
@@ -495,10 +516,10 @@ def shift_match(waves, alignment_channel,
         realigned_w = np.concatenate([w[-shift:,:], w[:-shift,:]], axis=0)
         if abs(relative_shift)>max_shift_allowed:
             realigned_w=realigned_w*np.nan
-        if plot_debug and i==0:
-            fig=imshow_cbar(template)
-            fig=imshow_cbar(w_closestchannels)
-            fig=imshow_cbar(xcorr_w_template)
+        # if plot_debug and i==0:
+        #     fig=imshow_cbar(template)
+        #     fig=imshow_cbar(w_closestchannels)
+        #     fig=imshow_cbar(xcorr_w_template)
         # store realigned_wave in array
         aligned_waves[i,:,:] = realigned_w
 
